/*
 DSPAM
 COPYRIGHT (C) 2004 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
                                                                                
*/

/*
   examples of libdspam implementation
   compile with:
   gcc -o example example.c -ldspam -L.libs -I. -DHAVE_CONFIG_H \
     -DDSPAM_HOME=/path/to/dspam-home
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libdspam.h>

#define USERNAME	"testuser"

int
main (int argc, char **argv)
{
  char buffer[1024];
  char *message = malloc (1);
  long len = 1;

  DSPAM_CTX *CTX;               	/* DSPAM Context */
  struct _ds_spam_signature SIG;        /* Example signature */

  /*
     Performs any driver-specific startup functions (such as initializing 
     internal lock tables).  Call this only once when your application
     starts
  */

  dspam_init_driver ();

  /* read in the message from stdin */
  message[0] = 0;
  while (fgets (buffer, sizeof (buffer), stdin) != NULL)
  {
    len += strlen (buffer);
    message = realloc (message, len);
    if (message == NULL)
    {
      fprintf (stderr, "out of memory!");
      exit (EXIT_FAILURE);
    }
    strcat (message, buffer);
  }

  /* MESSAGE PROCESSING */

  /* initialize dspam 

    Operating Modes:

    DSM_PROCESS:	Process message
    DSM_CLASSIFY:	Classify message only (do not write changes)

    Flags:

    DSF_CHAINED		Use Chained Tokens
    DSF_SIGNATURE	Signature Mode (Use a signature)
    DSF_NOISE		Use Bayesian Noise Reduction
    DSF_WHITELIST	Use Automatic Whitelisting
    DSF_COPYBACK	Copy back the decoded message

    Training Modes:

    DST_TEFT		Train Everything
    DST_TOE		Train-on-Error
    DST_TUM		Train-until-Mature

    Classifications:

    Used to tell libdspam the message has already been classified, and should
    be processed in such a way that the result will match the classification.

    DSR_ISSPAM		Message is spam (learn as spam)
    DSR_ISINNOCENT	Message is innocent (learn as innocent)
    DSR_NONE		No predetermined classification (classify message)

    Sources:

    Used to tell libdspam the source of the specified classification (if any).

    DSS_ERROR		Misclassification by libdspam
    DSS_CORPUS		Corpusfed message
    DSS_INOCULATION	Message inoculation
    DSS_NONE		No classification source (use only with DSR_NONE)

    NOTE: When using DSS_ERROR, a DSPAM signature should be provided, OR
          the original message in PRISTINE form (without any DSPAM headers and
          with the original message's headers).

  */ 

   /* --------------------------- EXAMPLE 1 ----------------------------*/
                      /* STANDARD INBOUND PROCESSING */

  /* In this example, DSF_SIGNATURE is specified to request that libdspam
     generate a signature, which can be stored internally for future retraining.
     This is useful if the original message in pristine form won't be available
     server-side should the user want to reclassify the message.
  */

  /* Initialize the DSPAM context */
  CTX = dspam_init (USERNAME, NULL, DSM_PROCESS,
                    DSF_CHAINED | DSF_SIGNATURE | DSF_NOISE | DSF_COPYBACK);

  if (CTX == NULL)
  {
    fprintf (stderr, "ERROR: dspam_init failed!\n");
    exit (EXIT_FAILURE);
  }

  /* Call DSPAM's processor with the message text */
  if (dspam_process (CTX, message) != 0)
  {
    fprintf (stderr, "ERROR: dspam_process failed");
    exit (EXIT_FAILURE);
  }

  /* Print processing results */
  printf ("Probability: %2.4f Confidence: %2.4f, Result: %s\n",
          CTX->probability,
          CTX->confidence,
          (CTX->result == DSR_ISSPAM) ? "Spam" : "Innocent");

  /* Manage signature */
  if (CTX->signature == NULL)
  {
    printf ("No signature provided\n");
  }
  else
  {
    /* Copy to a safe place */

    SIG.data = malloc (CTX->signature->length);
    if (SIG.data != NULL)
      memcpy (SIG.data, CTX->signature->data, CTX->signature->length);
  }
  SIG.length = CTX->signature->length;

  /* Destroy the context */

  /* CTX->message is an in/out variable, and should be destroyed whenever 
     you're finished using it, or if you're not going to use it.  Yes this
     is ugly. */

  _ds_destroy_message (CTX->message);

  if (dspam_destroy(CTX) != 0)
  {
    fprintf (stderr, "ERROR: dspam_destroy failed!");
    exit (EXIT_FAILURE);
  }

   /* --------------------------- EXAMPLE 2 ----------------------------*/
                 /* SPAM REPORTING (AS MISCLASSIFICATION) */

  /* We call everything just like before, with these exceptions:
     - We set the classification to DSR_ISSPAM
     - We set the source to DSS_ERROR

     This example will use the original message in pristine form instead of
     a signature.  See the next example (false positives) for an example using
     a signature.
  */

  /* Initialize the DSPAM context */
  CTX = dspam_init (USERNAME, NULL, DSM_PROCESS, DSF_CHAINED);
  if (CTX == NULL)
  {
    fprintf (stderr, "ERROR: dspam_init failed!\n");
    exit (EXIT_FAILURE);
  }

  /* Set up the context for error correction as spam */
  CTX->classification = DSR_ISSPAM;
  CTX->source         = DSS_ERROR;

  /* Call DSPAM */
  if (dspam_process(CTX, message) != 0)
  {
    fprintf (stderr, "ERROR: dspam_process failed\n");
    exit (EXIT_FAILURE);
  }

  /* Destroy the context */

  _ds_destroy_message (CTX->message);
  if (dspam_destroy (CTX) != 0)
  {
    fprintf (stderr, "ERROR: dspam_destroy failed!");
    exit (EXIT_FAILURE);
  }

  printf("Spam retrained successfully.\n");

   /* --------------------------- EXAMPLE 3 ----------------------------*/
                        /* FALSE POSITIVE REPORTING */

  /* Here we submit the message's signature for retraining as a false positive.
     We make the following changes from our original example: 
     - We set the classification to DSR_ISINNOCENT
     - We set the source to DSS_ERROR
     - We attach the binary signature to the context, and pass in a NULL text
  */

  /* Initialize DSPAM context */

  CTX = dspam_init (USERNAME, NULL,  DSM_PROCESS,
                    DSF_CHAINED | DSF_SIGNATURE);
  if (CTX == NULL)
  {
    fprintf (stderr, "ERROR: dspam_init failed!\n");
    exit (EXIT_FAILURE);
  }

  /* Set up the context for error correction as innocent */
  CTX->classification = DSR_ISINNOCENT;
  CTX->source         = DSS_ERROR;

  /* Attach the signature to the context */
  CTX->signature = &SIG;

  /* Call DSPAM */
  if (dspam_process(CTX, NULL) != 0)
  {
    fprintf (stderr, "ERROR: dspam_process failed\n");
    exit (EXIT_FAILURE);
  }

  /* Destroy the context */

  _ds_destroy_message (CTX->message);
  if (dspam_destroy (CTX) != 0)
  {
    fprintf (stderr, "ERROR: dspam_destroy failed!");
    exit (EXIT_FAILURE);
  }

  printf("False positive retrained successfully.\n");

  /* Performs any driver-specific shutdown functions */
  dspam_shutdown_driver ();

  exit (EXIT_SUCCESS);
}

